import re
from functools import partial

COLOURS = ("black", "red", "green", "yellow", "blue", "magenta", "cyan", "white")
STYLES = (
    "bold",
    "faint",
    "italic",
    "underline",
    "blink",
    "blink2",
    "negative",
    "concealed",
    "crossed",
)


class ColoursException(Exception):
    pass


def colour(s, foreground: str = None, background: str = None, style: str = None) -> str:
    sgr = []

    if foreground:
        if foreground in COLOURS:
            sgr.append(str(30 + COLOURS.index(foreground)))
        elif isinstance(foreground, int) and foreground in range(0, 256):
            sgr.append(f"38;5;{foreground}")
        else:
            raise ColoursException(f'Invalid colour "{foreground}"')

    if background:
        if background in COLOURS:
            sgr.append(str(40 + COLOURS.index(background)))
        elif isinstance(background, int) and background in range(0, 256):
            sgr.append(f"48;5;{background}")
        else:
            raise ColoursException(f'Invalid colour "{background}"')

    if style:
        for st in style.split("+"):
            if st in STYLES:
                sgr.append(str(1 + STYLES.index(st)))
            else:
                raise ColoursException(f'Invalid style "{st}"')

    if sgr:
        prefix = r"\x1b[" + ";".join(sgr) + "m"
        suffix = r"\x1b[0m"
        return f"{prefix}{s}{suffix}"
    else:
        return s


def strip_color(s: str) -> str:
    return re.sub(r"\x1b\[.+?m", "", s)


# Foreground shortcuts
black = partial(colour, foreground="black")
red = partial(colour, foreground="red")
green = partial(colour, foreground="green")
yellow = partial(colour, foreground="yellow")
blue = partial(colour, foreground="blue")
magenta = partial(colour, foreground="magenta")
cyan = partial(colour, foreground="cyan")
white = partial(colour, foreground="white")

# Style shortcuts
bold = partial(colour, style="bold")
faint = partial(colour, style="faint")
italic = partial(colour, style="italic")
underline = partial(colour, style="underline")
blink = partial(colour, style="blink")
blink2 = partial(colour, style="blink2")
negative = partial(colour, style="negative")
concealed = partial(colour, style="concealed")
crossed = partial(colour, style="crossed")





def iprint(
        imput: str,
        autoclose: bool = True,
        autoprint: str = True,
        clean: bool = False,
        **kwargs
):
    """
    Utility function that provides interactive prints with unicode activated curses manipulation

    Uses a concatenation formatting syntax, for example:

      >>> from inprint import iprint
      >>> iprint('[b]This text will print in blue.[/]')
      >>> iprint('[*][r]This text will print in bold and red.[/]')
      >>> iprint('[*][r]Bold red text: [/][y]Non-bold yellow text, [b]blue text, [*]bold blue text.[/]')

    :param imput: str: Input stream to format and print
    :param autoclose: bool: Defaults to `True`: Automatically append the curses termination byte
    :param autoprint: bool: Defaults to `True`: Automatically pass the formatted string to `print()`
    :param clean: bool: Defaults to `False`: If set to True, all color tags will be removed to clean the string.
    :param kwargs: dict: All remaining arguments transparently passed on to the builtin. See parameters for `print()`
    :return: str: The formatted string result
    """
    colours = {
        "black": ("d", "\x1b[30m"),
        "red": ("r", "\x1b[31m"),
        "green": ("g", "\x1b[32m"),
        "yellow": ("y", "\x1b[33m"),
        "blue": ("b", "\x1b[34m"),
        "magenta": ("m", "\x1b[35m"),
        "cyan": ("c", "\x1b[36m"),
        "white": ("w", "\x1b[37m"),
    }
    bold_colours = {
        "black": ("D", "\x1b[1m\x1b[30m"),
        "red": ("R", "\x1b[1m\x1b[31m"),
        "green": ("G", "\x1b[1m\x1b[32m"),
        "yellow": ("Y", "\x1b[1m\x1b[33m"),
        "blue": ("B", "\x1b[1m\x1b[34m"),
        "magenta": ("M", "\x1b[1m\x1b[35m"),
        "cyan": ("C", "\x1b[1m\x1b[36m"),
        "white": ("W", "\x1b[1m\x1b[37m"),
    }
    styles = {
        "bold": ("*", "\x1b[1m"),
        "faint": ("~", "\x1b[2m"),
        "italic": ("i", "\x1b[3m"),
        "underline": ("_", "\x1b[4m"),
        "blink": ("bl", "\x1b[5m"),
        "blink2": ("bl2", "\x1b[6m"),
        "negative": ("neg", "\x1b[7m"),
        "concealed": ("?", "\x1b[8m"),
        "crossed": ("-", "\x1b[9m"),
        "reset": ("/", "\x1b[0m"),
    }
    for name, colour in bold_colours.items():
        imput = imput.replace(f"[{colour[0]}]", colour[1] if not clean else "")
    for name, colour in colours.items():
        imput = imput.replace(f"[{colour[0]}]", colour[1] if not clean else "")
    for name, style in styles.items():
        imput = imput.replace(f"[{style[0]}]", style[1] if not clean else "")
    if autoclose and not clean:
        imput += styles["reset"][1]
    if autoprint:
        print(imput, **kwargs)
    return imput
